'#################################################
'# 
'# fbTimer: Replacement for Timer with several improvements and additions
'#          including optional fully threaded scheduling and event callback
'# 
'# Version: 0.1
'# 
'# Optional requirement for Event -scheduler, -timeouts, -callbacks:
'#     fbList.bi - Available at www.JatteGames.com
'# 
'# Official site: www.JatteGames.com
'# 
'# fbTimer is Copyright (c) 2023-2024 Johannes "Jattenalle" Pihl, all rights reserved
'# 
'# 	This software is provided 'as-is', without any express or implied
'# 	warranty. In no event will the authors be held liable for any damages
'# 	arising from the use of this software.
'# 	
'# 	Permission is granted to anyone to use this software for any purpose,
'# 	including commercial applications, and to alter it and redistribute it
'# 	freely, subject to the following restrictions:
'# 	
'# 	1. The origin of this software must not be misrepresented; you must not
'# 	   claim that you wrote the original software. If you use this software
'# 	   in a product, an acknowledgment in the product documentation would be
'# 	   appreciated but is not required.
'# 	2. Altered source versions must be plainly marked as such, and must not be
'# 	   misrepresented as being the original software.
'# 	3. This notice may not be removed or altered from any source distribution.
'# 
'#################################################
'# 
'# Known issues:
'# 	  None
'# 
'#################################################
'# 
'# Event callback sub structure:
'# 	  sub(byref Event as TimerEvent)
'# 
'#################################################
'# 
'# Use #DEFINE FBTIMER_NOLIST before you include fbTimer to not require
'#  FBLIST. Note that this will disable fbTimer events and scheduling!
'# Example:
'#    #DEFINE FBTIMER_NOLIST
'#    #INCLUDE ONCE "fbTimer.bi"
'# 
'#################################################
'# 
'# Use #DEFINE FBTIMER_INCLUDE_HEADER_ONLY before you include fbTimer to only
'#  include the declares and definitions without the methods and procedures
'#  themselves. Can be combined with FBTIMER_NOLIST.
'# Example:
'#    #DEFINE FBTIMER_INCLUDE_HEADER_ONLY
'#    #INCLUDE ONCE "fbTimer.bi"
'# 
'#################################################

#ifdef FBTIMER_INC
	#cmdline "-maxerr 1"
	#error  "FBTIMER is already included! It can only be included once to ensure #defines are handled properly. Please verify that you only #include FBTIMER once."
#else
	#define FBTIMER_INC

	#define FBTIMER_VERSION			"0.1"
	
	#if __FB_VERSION__ < "1.10.1"
		#cmdline "-maxerr 1"
		#print "You are using FBC version "
		#print __FB_VERSION__
		#print "FBTIMER requires FreeBASIC 1.10.1 or later!"
		#error "Update FreeBASIC!"
	#endif

	#ifndef FBTIMER_NOLIST
		#ifndef FBLIST_INC
			#cmdline "-maxerr 1"
			#error "FBLIST not included! Include FBLIST before FBTIMER or use #DEFINE FBTIMER_NOLIST to disable FBTIMER event features - FBLIST is Available at www.JatteGames.com"
		#else
			#if FBLIST_VERSION < "0.3"
				#cmdline "-maxerr 1"
				#error "FBTIMER requires FBLIST version 0.3 or greater! - FBLIST is Available at www.JatteGames.com"
			#endif
			#ifdef __FBLIST_PP_DISABLEMACROS
				#cmdline "-maxerr 1"
				#error "FBTIMER requires FBLIST macros! Please undefine __FBLIST_PP_DISABLEMACROS before including FBLIST."
			#endif
		#endif
	#endif
	
	
	type TIMER_ID as ulongint
	const as integer TIMER_ONCE				=	1
	const as integer TIMER_FOREVER			=	-108011556
	const as integer TIMER_CALLBACKS_ALL	=	-108011556
	const as TIMER_ID TIMER_ALL				=	(2ull shl 60ull) + 1ull
	dim shared as longint __TIMERCORE_ACCURACY
	
	declare function __TIMERCORE_TIMESTAMP() as ulongint
	
	#ifndef FBTIMER_INCLUDE_HEADER_ONLY
		#ifdef __FB_WIN32__
		'__FB_WIN32__
		'__FB_LINUX__
			#ifndef _INC_WINDOWS
				#include once "windows.bi"
			#endif
			'#################################################
			'# 
			'# Default TIMER also uses the QPC but casts the result to a double,
			'# which does not hold enough precision to allow the actual full range of values
			'# returned by the QPC for systems with high (Months+) uptimes
			'# 
			'#################################################
			function __TIMERCORE_TIMESTAMP() as ulongint
				static as LARGE_INTEGER ptStart
				static as LARGE_INTEGER ptDivisor
				static as LARGE_INTEGER ptCounter
				if ptDivisor.QuadPart = 0 then
					QueryPerformanceCounter(@ptStart)
					dim as integer retVal = QueryPerformanceFrequency(@ptDivisor)
					if ptDivisor.QuadPart = 0 or retVal = 0 then	'// This will happen if ptDivisor is NOT 8byte aligned
						dim as integer wErrC = GetLastError()
						dim as integer f = freefile
						open cons for output as #f
						print #f, "ERROR: FBTIMER QueryPerformanceFrequency() returned 0 with error message 0x"& hex(wErrC) &"."
						if cint(@ptDivisor) mod 8 <> 0 then
							print #f, "        Possible cause? ptDivisor (0x"& hex(@ptDivisor) &") is not 8byte aligned due to compiler variable byte alignment settings!"
						end if
						close #f
					'	error -1	'// At this point we can't do anything but die
						end
					end if
					__TIMERCORE_ACCURACY = ptDivisor.QuadPart
				end if
				QueryPerformanceCounter(@ptCounter)
				
				return (((ptCounter.QuadPart - ptStart.QuadPart) / ptDivisor.QuadPart) * 10000000LL)
			end function
		#else
			'// Fall back to default fb method, which is less precise due to casting to double and losing precision
			__TIMERCORE_ACCURACY = 0 '// Impossible to know because the result <double> accuracy is dependant on system uptime etc
			function __TIMERCORE_TIMESTAMP() as ulongint
				static as longint ptStart
				if ptStart = 0 then
					ptStart = (TIMER * 10000000ll) - 1
				end if
				return (TIMER * 10000000ll) - ptStart
			end function
		#endif
	#endif
	
	type __TimerVal
		declare function s as double option("SSE")
		declare function ms as ulongint
		declare function us as ulongint
		declare function dblms as double option("SSE")
		declare function _raw as ulongint
		declare operator cast() as double option("SSE")
		declare operator cast() as string
		pCreation as ulongint ptr
	end type
	function __TimerVal.s as double
		static as ulongint ts
		ts = __TIMERCORE_TIMESTAMP
		if ts > *this.pCreation then
			return (ts - *this.pCreation) / 10000000.0
		end if
		return 0
	end function
	function __TimerVal.ms as ulongint
		static as ulongint ts
		ts = __TIMERCORE_TIMESTAMP
		if ts > *this.pCreation then
			return (ts - *this.pCreation) / 10000
		end if
		return 0
	end function
	function __TimerVal.us as ulongint
		static as ulongint ts
		ts = __TIMERCORE_TIMESTAMP
		if ts > *this.pCreation then
			return (ts - *this.pCreation) / 10
		end if
		return 0
	end function
	function __TimerVal._raw as ulongint
		static as ulongint ts
		ts = __TIMERCORE_TIMESTAMP
		if ts > *this.pCreation then
			return (ts - *this.pCreation)
		end if
		return 0
	end function
	function __TimerVal.dblms as double
		static as ulongint ts
		ts = __TIMERCORE_TIMESTAMP
		if ts > *this.pCreation then
			return (ts - *this.pCreation) / 10000.0
		end if
		return 0
	end function
	operator __TimerVal.cast() as double
		static as ulongint ts
		ts = __TIMERCORE_TIMESTAMP
		if ts > *this.pCreation then
			return (ts - *this.pCreation) / 10000000.0
		end if
		return 0
	end operator
	operator __TimerVal.cast() as string
		static as ulongint ts
		ts = __TIMERCORE_TIMESTAMP
		if ts > *this.pCreation then
			return str((ts - *this.pCreation) / 10000000.0)
		end if
		return "0"
	end operator
	
	#ifndef FBTIMER_NOLIST
		type _____t_Timer as t_Timer '// Forward definition
		type TimerEvent
			Lifetime as __TimerVal
			Creation as ulongint
			Timeout as double
			Count as longint
			Remain as integer
			Repeats as integer
			UserData as any ptr
			TID as TIMER_ID
			declare constructor()
			_pParent as _____t_Timer ptr
		end type
		constructor TimerEvent()
			this.Lifetime.pCreation = @this.Creation
		end constructor
		type _SchedulerEvent
			Active as integer
			Callback as sub(byref Event as TimerEvent)
			e as TimerEvent
		end type
		typelist _SchedulerEventList holds _SchedulerEvent
	#endif

	type __TimerInterval
		Interval as ulongint
		Touch as ulongint
	end type
	type t_Timer extends __TimerVal
		#ifndef FBTIMER_NOLIST
			declare function Timeout(byval duration as double, byval callback as sub(byref Event as TimerEvent), byval userData as any ptr = 0) as integer
			declare function Timeout(byval duration as double, byval callback as sub(byref Event as TimerEvent), byval userData as integer) as integer
			declare function Schedule(byval duration as double, byval callback as sub(byref Event as TimerEvent), byval Repeats as integer = TIMER_FOREVER, byval userData as any ptr = 0) as integer
			declare function Schedule(byval duration as double, byval callback as sub(byref Event as TimerEvent), byval Repeats as integer = TIMER_FOREVER, byval userData as integer) as integer
			declare function Callbacks(byval nMaxCallbacks as integer = TIMER_CALLBACKS_ALL) as integer
			declare function Events() as integer
		#endif
		declare function Interval(byval _interval as double) as integer
		declare function Interval() as integer
		declare function GetInterval() as double option("SSE")
		Delta as double
		
		declare sub Start()
		declare sub Pause(byval tid as TIMER_ID = TIMER_ALL)
		declare sub Resume(byval tid as TIMER_ID = TIMER_ALL)
		declare sub Remove(byval tid as TIMER_ID = TIMER_ALL)
		declare sub Reset(byval tid as TIMER_ID = TIMER_ALL)
		
		declare function isRunning(byval tid as TIMER_ID = TIMER_ALL) as integer
		
		declare function ID() as TIMER_ID
		
		Runtime as __TimerVal
		
		declare operator cast() as double option("SSE")
		declare operator cast() as string
		
		declare constructor()
		declare destructor()
		
		declare function Accuracy() as ulongint
		
		declare sub sleep(byval d as double)
		private:
			#ifndef FBTIMER_NOLIST
				_EventList as _SchedulerEventList
			#endif
			_TID as TIMER_ID
			_Creation as ulongint
			_TimerInstanceMutEx as any ptr
			_pInterval as __TimerInterval
			_Active as integer
			_InactiveDuration as ulongint
	end type

	#ifndef FBTIMER_INCLUDE_HEADER_ONLY
		function __TIMERCORE_TID() as TIMER_ID
			static as any ptr TIDMutex
			static as ulongint __TIDC
			if TIDMutex = 0 then
				TIDMutex = mutexcreate()
			end if
			mutexlock TIDMutex
			__TIDC += 1
			if __TIDC >= (2ull shl 60) then
				__TIDC -= (2ull shl 60)
			end if
			if __TIDC = 0 then
				__TIDC += 1
			end if
			mutexunlock TIDMutex
			return __TIDC
		end function
		dim shared as ulongint __TIMERCORE_CREATION
		__TIMERCORE_CREATION = __TIMERCORE_TIMESTAMP
		constructor t_Timer()
			this._TID						=	__TIMERCORE_TID
			this._TimerInstanceMutEx		=	mutexcreate
			this.pCreation					=	@this._Creation
			this.Runtime.pCreation			=	@__TIMERCORE_CREATION
			this._Creation					=	__TIMERCORE_TIMESTAMP
			this._Active					=	TRUE
		end constructor
		destructor t_Timer()
			mutexdestroy this._TimerInstanceMutEx
			this._TimerInstanceMutEx = 0
		end destructor
		operator t_Timer.cast() as double
			return this.s
		end operator
		operator t_Timer.cast() as string
			return str(this.s)
		end operator
		
		function t_Timer.Accuracy() as ulongint
			return __TIMERCORE_ACCURACY
		end function
		
		function t_Timer.ID() as TIMER_ID
			return this._TID
		end function
		
		#ifndef FBTIMER_NOLIST
			function t_Timer.Callbacks(byval nMaxCallbacks as integer = TIMER_CALLBACKS_ALL) as integer
				dim as integer nEvents
				dim as ulongint filter = __TIMERCORE_TIMESTAMP
				
				if This._EventList.Length then
					mutexlock This._TimerInstanceMutEx
					for e as integer = 0 to This._EventList.Length -1
						if This._EventList.Length <= 0 then '// There's nothing here anymore
							exit for
						end if
						if e >= This._EventList.Length then '// The list length can change
							exit for
						end if
						
						if This._EventList[e].Active then
							if This._EventList[e].e.Creation > 0 and This._EventList[e].e.Creation < filter then
								if This._EventList[e].e.Lifetime.dblms >= This._EventList[e].e.Timeout then
									if This._EventList[e].e.Repeats <> TIMER_FOREVER then
										This._EventList[e].e.Remain -= 1
									end if
									This._EventList[e].e.Count += 1
									
									'// We have to unlock the mutex while calling the callback in case it manipulates the same timer
									mutexunlock This._TimerInstanceMutEx
									This._EventList[e].Callback(This._EventList[e].e)
									mutexlock This._TimerInstanceMutEx
									
									nEvents += 1
									if This._EventList.GetPtr(e) <> 0 then '// We check again on the off-chance that the callback deleted the variable
										This._EventList[e].e.Creation = __TIMERCORE_TIMESTAMP
										if This._EventList[e].e.Repeats <> TIMER_FOREVER then
											if This._EventList[e].e.Remain <= 0 then
												This._EventList.Remove(e)
												e-=1 '// Step back since we removed it and now the list has shifted left
											end if
										end if
									end if
								end if
							end if
						end if
						if nMaxCallbacks <> TIMER_CALLBACKS_ALL andalso nEvents > nMaxCallbacks then
							exit for
						end if
					next
					mutexunlock This._TimerInstanceMutEx
				end if
				return nEvents
			end function
			function t_Timer.Events() as integer
				return This._EventList.Length
			end function
			function t_Timer.Timeout(byval duration as double, byval callback as sub(byref Event as TimerEvent), byval userData as integer) as integer
				return this.Schedule(duration, callback, 1, cptr(any ptr, userData))
			end function
			function t_Timer.Schedule(byval duration as double, byval callback as sub(byref Event as TimerEvent), byval Repeats as integer, byval userData as integer) as integer
				return this.Schedule(duration, callback, Repeats, cptr(any ptr, userData))
			end function
			function t_Timer.Timeout(byval duration as double, byval callback as sub(byref Event as TimerEvent), byval userData as any ptr = 0) as integer
				return this.Schedule(duration, callback, 1, userData)
			end function
			function t_Timer.Schedule(byval duration as double, byval callback as sub(byref Event as TimerEvent), byval Repeats as integer, byval userData as any ptr = 0) as integer
				mutexlock this._TimerInstanceMutEx
				dim as integer index = this._EventList.Insert(new _SchedulerEvent)
				dim as _SchedulerEvent ptr sep = this._EventList.GetPtr(index)
				
				dim as TIMER_ID tid
				if index > -1 then
					tid = __TIMERCORE_TID
					sep->Callback		=	callback
					sep->e.TID			=	tid
					sep->e._pParent		=	@this
					sep->e.Timeout		=	duration
					sep->e.Creation		=	__TIMERCORE_TIMESTAMP
					if sep->e.Creation = 0 then
						sep->e.Creation = 1
					end if
					sep->e.Remain		=	Repeats
					sep->e.Repeats		=	Repeats
					sep->e.UserData		=	userData
					sep->Active			=	TRUE
				end if
				mutexunlock this._TimerInstanceMutEx
				return tid
			end function
		#endif
		
		sub t_Timer.Sleep(byval duration as double)
			dim as ulongint n = __TIMERCORE_TIMESTAMP
			dim as longint d = duration * 10000
			if d <= 0 then
				exit sub
			end if
			n += d
			do
				if __TIMERCORE_TIMESTAMP >= n then
					exit do
				end if
			loop
		end sub
		function t_Timer.GetInterval() as double
			return this._pInterval.Interval / 10000
		end function
		function t_Timer.Interval(byval _interval as double) as integer
			if _interval < 0 then
				return FALSE
			end if
			if _interval > 0 then
				this._pInterval.Interval	=	_interval * 10000
				this._pInterval.Touch		=	0
				return TRUE
			end if
			return FALSE
		end function
		function t_Timer.Interval() as integer
			static as ulongint ts
			if this._Active then
				if this._pInterval.Interval then
					ts = __TIMERCORE_TIMESTAMP
					if this._pInterval.Touch = 0 then
						this._pInterval.Touch = ts
					elseif ts >= this._pInterval.Touch + this._pInterval.Interval then
						this.Delta				=	(ts - this._pInterval.Touch) / 10000
						this._pInterval.Touch	+=	this._pInterval.Interval
						return TRUE
					end if
				end if
			end if
			return FALSE
		end function
		sub t_Timer.Start()
			this.pCreation			=	@this._Creation
			this._pInterval.Touch	=	0
			this._Creation			=	__TIMERCORE_TIMESTAMP
			this._Active			=	TRUE
		end sub
		sub t_Timer.Reset(byval tid as TIMER_ID = TIMER_ALL)
			mutexlock this._TimerInstanceMutEx
			#ifndef FBTIMER_NOLIST
				dim as integer nevents = this._EventList.Length
				for e as integer = 0 to nevents - 1
					if tid = TIMER_ALL or tid = this._EventList[e].e.TID then
						this._EventList[e].e.Creation = __TIMERCORE_TIMESTAMP
					end if
				next
			#endif
			this._pInterval.Touch	=	0
			mutexunlock this._TimerInstanceMutEx
		end sub
		sub t_Timer.Pause(byval tid as TIMER_ID = TIMER_ALL)
			mutexlock this._TimerInstanceMutEx
			#ifndef FBTIMER_NOLIST
			dim as integer nevents = this._EventList.Length
			for e as integer = 0 to nevents - 1
				if tid = TIMER_ALL or tid = this._EventList[e].e.TID then
					this._EventList[e].Active = FALSE
				end if
			next
			#endif
			if this._Active then
				this._InactiveDuration	=	__TIMERCORE_TIMESTAMP
			end if
			this._Active	=	FALSE
			mutexunlock this._TimerInstanceMutEx
		end sub
		sub t_Timer.Resume(byval tid as TIMER_ID = TIMER_ALL)
			mutexlock this._TimerInstanceMutEx
			dim as ulongint stamp = __TIMERCORE_TIMESTAMP
			#ifndef FBTIMER_NOLIST
				dim as integer nevents = this._EventList.Length
				for e as integer = 0 to nevents - 1
					if tid = TIMER_ALL or tid = this._EventList[e].e.TID then
						this._EventList[e].Active = TRUE
						if this._InactiveDuration < stamp then
							this._EventList[e].e.Creation += (stamp - this._InactiveDuration)
						end if
					end if
				next
			#endif
			this._Active	=	TRUE
			if this._pInterval.Interval then
				if this._InactiveDuration < stamp then
					this._pInterval.Touch	+=	(stamp - this._InactiveDuration)
				else
					this._pInterval.Touch	=	0
				end if
			end if
			mutexunlock this._TimerInstanceMutEx
		end sub
		sub t_Timer.Remove(byval tid as TIMER_ID = TIMER_ALL)
			mutexlock this._TimerInstanceMutEx
			#ifndef FBTIMER_NOLIST
				dim as integer nevents = this._EventList.Length
				for e as integer = 0 to nevents - 1
					if e >= this._EventList.Length then '// List length can change
						exit for
					end if
					if this._EventList.Length = 0 then '// List length can change
						exit for
					end if
					if tid = TIMER_ALL or tid = this._EventList[e].e.TID then
						deallocate this._EventList.RawListMemory[e]
						this._EventList.RawListMemory[e] = 0
						this._EventList.Remove(e)
						e-=1 '// Step back
					end if
				next
			#endif
			mutexunlock this._TimerInstanceMutEx
		end sub
		function t_Timer.isRunning(byval tid as TIMER_ID = TIMER_ALL) as integer
			mutexlock this._TimerInstanceMutEx
			#ifndef FBTIMER_NOLIST
				dim as integer nevents = this._EventList.Length
				for e as integer = 0 to nevents - 1
					if tid = TIMER_ALL or tid = this._EventList[e].e.TID then
						if this._EventList[e].Active then
							mutexunlock this._TimerInstanceMutEx
							return TRUE
						end if
					end if
				next
			#endif
			mutexunlock this._TimerInstanceMutEx
			if this._Active then
				return TRUE
			end if
			return FALSE
		end function
	#endif
#endif